class Log:
    def __init__(self):
        self.entries = []
        self.js = None
        self.max_entries = 100

    def set_js(self, js):
        """
        Set a callback to JavaScript code for communication.
        :param js: Callback function set by JavaScript
        :return: Initial log contents
        """
        # todo ; rename function to set_js_callback to be more descriptive
        self.js = js
        self.read_all()

    def write(self, entry, echo=True):
        self.entries.append(entry)
        while len(self.entries) > self.max_entries:
            self.entries.pop(0)
        if echo:
            print(entry.strip())
        if self.js:
            # update js log
            self.read_all()

    def read_all(self):
        entries = "\n".join(self.entries).strip()
        if self.js:
            self.js.Call(entries)
        return entries



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Slipstream - The most informative Home-media backup solution.
Copyright (C) 2020 PHOENiX

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

~~~

Various small simple helper functions to do a quick task while not
being specific enough to be in a class.
"""
import builtins as g
import queue
import subprocess
import threading

import pslipstream.cfg as cfg
from pycdlib import PyCdlib

if cfg.windows:
    from win32 import win32api, win32file


def list_devices():
    """
    Lists all devices provided by lsscsi
    """
    if cfg.windows:
        drives = [
            rf"\\.\{d[:-1]}" for d in win32api.GetLogicalDriveStrings().split('\x00')[:-1]
            if win32file.GetDriveType(d) == win32file.DRIVE_CDROM
        ]
        print(drives)
        drives = [{
            "loc": d,
            "volid": get_volume_id(d)
        } for d in drives]
        return drives
    if cfg.linux:
        lsscsi = subprocess.check_output(["lsscsi"]).decode().splitlines()
        lsscsi = [x[9:].strip() for x in lsscsi]

        lsscsi = [[x for x in scsi.split(" ") if x] for scsi in lsscsi]
        lsscsi = [{
            "type": scsi[0],
            "make": scsi[1],
            "model": " ".join([scsi[2]] if len(scsi) == 5 else scsi[2:(len(scsi) - 2)]),
            "fwver": scsi[-2],
            "loc": scsi[-1],
            "volid": get_volume_id(scsi[-1])
        } for scsi in lsscsi if scsi[0] not in ["disk"]]
        return lsscsi


def get_volume_id(device):
    """
    Get the Volume Identifier for a device

    Returns None if there's no disc inserted
    """
    cdlib = PyCdlib()
    try:
        cdlib.open(device, "rb")
    except OSError as e:
        # noinspection SpellCheckingInspection
        if "[Errno 123]" in str(e):
            # no disc inserted
            g.LOG.write(f"Device {device} has no disc inserted.")
            return None
        # noinspection SpellCheckingInspection
        if "[Errno 5]" in str(e):
            # Input/output error
            g.LOG.write(f"Device {device} had an I/O error.")
            return "! Error occurred reading disc..."
        raise
    volume_id = cdlib.pvds[0].volume_identifier.decode().strip()
    g.LOG.write(f"Device {device} has disc labeled \"{volume_id}\".")
    return volume_id


def get_device_list(js):
    js.Call(sorted(list_devices(), key=lambda d: d["volid"] or "", reverse=True))


def asynchronous(f, daemon=True, auto_start=False):
    def wrapped_f(q, *args, **kwargs):
        """
        this function calls the decorated function and puts the
        result in a queue
        """
        ret = f(*args, **kwargs)
        q.put(ret)

    def wrap(*args, **kwargs):
        """
        this is the function returned from the decorator. It fires off
        wrapped_f in a new thread and returns the thread object with
        the result queue attached
        """

        q = queue.Queue()

        t = threading.Thread(target=wrapped_f, args=(q,) + args, kwargs=kwargs)
        t.daemon = daemon
        if auto_start:
            t.start()
        t.result_queue = q
        return t

    return wrap


def asynchronous_auto(f):
    return asynchronous(f, auto_start=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Slipstream - The most informative Home-media backup solution.
Copyright (C) 2020 PHOENiX

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import sys
import tkinter as tk

from cefpython3 import cefpython as cef

import pslipstream.cfg as cfg
import pslipstream.exceptions as exceptions
from pslipstream.gui.browser_frame import BrowserFrame


class Gui(tk.Frame):

    def on_root_configure(self, _):
        if self.browser_frame:
            self.browser_frame.on_root_configure()

    def on_configure(self, event):
        if self.browser_frame:
            width = event.width
            height = event.height
            self.browser_frame.on_mainframe_configure(width, height)

    def on_close(self):
        if self.browser_frame:
            self.browser_frame.on_root_close()
        self.master.destroy()
        cef.Shutdown()

    def __init__(self, url, icon, js_bindings=None):
        self.js_bindings = js_bindings
        self.browser_frame = None
        self.master = None
        # Assure Tkinter version is supported
        if tk.TkVersion <= 8.5:
            raise exceptions.TkinterVersionError()
        # noinspection SpellCheckingInspection
        # tell cef about uncaught exceptions, will shutdown
        sys.excepthook = cef.ExceptHook
        # Create Window
        root = tk.Tk()
        root.geometry(cfg.min_size)
        root.update()
        root.minsize(root.winfo_width(), root.winfo_height())
        tk.Grid.rowconfigure(root, 0, weight=1)
        tk.Grid.columnconfigure(root, 0, weight=1)
        tk.Frame.__init__(self, root)
        # Configure Window
        # noinspection PyUnresolvedReferences
        self.master.title(f"{cfg.title} v{cfg.version}")
        # noinspection PyUnresolvedReferences
        self.master.protocol("WM_DELETE_WINDOW", self.on_close)
        # noinspection PyUnresolvedReferences
        self.master.bind("<Configure>", self.on_root_configure)
        # noinspection PyUnresolvedReferences
        # noinspection PyProtectedMember
        self.master.call(
            "wm",
            "iconphoto",
            self.master._w,
            tk.PhotoImage(file=icon)
        )
        self.bind("<Configure>", self.on_configure)
        # BrowserFrame
        self.browser_frame = BrowserFrame(self, url=url, js_bindings=self.js_bindings)
        self.browser_frame.grid(row=0, column=0, sticky=(tk.N + tk.S + tk.E + tk.W))
        self.browser = self.browser_frame.browser
        tk.Grid.rowconfigure(self, 0, weight=1)
        tk.Grid.columnconfigure(self, 0, weight=1)
        # Pack MainFrame
        self.pack(fill=tk.BOTH, expand=tk.YES)
        # noinspection SpellCheckingInspection
        # Initialize CEF
        if not cef.Initialize(settings={
            "debug": False,
            "command_line_args_disabled": True,
            "context_menu": {  # we use a custom context menu all the time, we cant have both
                "enabled": False,
                "navigation": False,
                "print": False,
                "view_source": False,
                "external_browser": False,
                "devtools": False
            },
            "cache_path": "",  # disable
            "ignore_certificate_errors": False,
            "downloads_enabled": False,
            "locale": "en-US",
            "log_severity": cef.LOGSEVERITY_DISABLE,
            "persist_session_cookies": False,
            "persist_user_preferences": False,
            "remote_debugging_port": -1,
            "product_version": f"Slipstream/{cfg.version}",  # user agent for the UI
            "background_color": 0xff202225
        }, switches={
            "no-proxy-server": "",  # avoid using ie set proxy, if they want to system-wide proxy, use vpn
            "allow-file-access-from-files": ""  # so we can use locally stored files for includes on main html file
        }):
            raise exceptions.CefInitializationError()



import sys

from setuptools import Command

from setup_commands import build


class DistCommand(Command):
    """Support setup.py build."""

    description = "Build the package."
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    @staticmethod
    def run():
        build()
        sys.exit()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Slipstream - The most informative Home-media backup solution.
Copyright (C) 2020 PHOENiX

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

~~~

Class file that handles all DVD operations including loading,
reading, seeking, backing up, and more.
"""

import builtins as g
import os
import sys
from datetime import datetime

import pycdlib
import rlapydvdid
from dateutil.tz import tzoffset
from pydvdcss.dvdcss import DvdCss
from tqdm import tqdm

import pslipstream.cfg as cfg
from pslipstream.exceptions import SlipstreamSeekError, SlipstreamDiscInUse, SlipstreamNoKeysObtained, \
    SlipstreamReadError
from pslipstream.helpers import asynchronous_auto


class Dvd:
    def __init__(self):
        self.dev = None
        self.ready = False
        self.cdlib = None
        self.dvdcss = None
        self.reader_position = 0
        self.vob_lba_offsets = []

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.dispose()

    def dispose(self):
        g.LOG.write(f"Disposing Dvd object...")
        if self.cdlib:
            self.cdlib.close()
        if self.dvdcss:
            self.dvdcss.dispose()
        self.__init__()  # reset everything
        g.PROGRESS.c.Call(0)

    @asynchronous_auto
    def open(self, dev, js=None):
        """
        Open the device as a DVD with pycdlib and libdvdcss.

        pycdlib will be used to identify and extract information.
        libdvdcss will be used for reading, writing, and decrypting.

        Raises SlipstreamDiscInUse if you try to load the same disc that's
        already opened. You can open a different disc without an exception as
        it will automatically dispose the current disc before opening.
        """
        if self.dvdcss or self.cdlib:
            if dev != self.dev:
                # dispose, and continue loading the new disc
                self.dispose()
            else:
                raise SlipstreamDiscInUse("The specified DVD device is already open in this instance.")
        self.dev = dev
        g.LOG.write(f"Opening {dev} as a DVD...")
        self.cdlib = pycdlib.PyCdlib()
        self.cdlib.open("\\\\.\\" + dev if cfg.windows else dev)
        g.LOG.write(f"Initialised pycdlib instance successfully...")
        self.dvdcss = DvdCss()
        self.dvdcss.open(dev)
        g.LOG.write(f"Initialised pydvdcss instance successfully...")
        self.ready = True
        g.LOG.write(f"DVD opened and ready...\n")
        if js:
            js.Call()

    def is_ready(self, js):
        """
        Simple function just to be able to check if this Dvd
        instance is ready to be used for reading and what not.

        Cefpython cannot export the Properties of Dvd, so this
        is the next best way of doing it.
        """
        js.Call(self.ready)

    @asynchronous_auto
    def compute_crc_id(self, js=None):
        """
        Get the CRC64 checksum known as the Media Player DVD ID.
        The algorithm used is the exact same one used by Microsoft's old Windows Media Center.
        """
        crc = str(rlapydvdid.compute(self.dev))
        g.LOG.write(f"Got CRC64 DVD ID: {crc}\n")
        if js:
            js.Call(crc)
        return crc

    @asynchronous_auto
    def get_primary_descriptor(self, js=None):
        """
        Get's and returns the Primary Volume Descriptor of the
        disc in a more accessible and parsed format.
        """
        pvd = self.cdlib.pvds[0]

        def date_convert(d):
            if not d.year:
                return None
            return datetime(
                year=d.year, month=d.month, day=d.dayofmonth,
                hour=d.hour, minute=d.minute, second=d.second, microsecond=d.hundredthsofsecond,
                # offset the timezone, since ISO's dates are offsets of GMT in 15 minute intervals, we
                # need to calculate that but in seconds to pass to `tzoffset`.
                tzinfo=tzoffset("GMT", (15 * d.gmtoffset) * 60)
            )

        pvd = {
            "version": pvd.version,
            "version_fs": pvd.file_structure_version,
            "flags": pvd.flags,
            "sector_size": pvd.log_block_size,
            "total_sectors": pvd.space_size,
            "size": pvd.log_block_size * pvd.space_size,
            "system_id": pvd.system_identifier.decode().strip() or None,
            "volume_id": pvd.volume_identifier.decode().strip() or None,
            "volume_set_id": pvd.volume_set_identifier.decode().strip() or None,
            "publisher_id": pvd.publisher_identifier.record().decode().strip() or None,
            "preparer_id": pvd.preparer_identifier.record().decode().strip() or None,
            "application_id": pvd.application_identifier.record().decode().strip() or None,
            "copyright_file_id": pvd.copyright_file_identifier.decode().strip() or None,
            "abstract_file_id": pvd.abstract_file_identifier.decode().strip() or None,
            "bibliographic_file_id": pvd.bibliographic_file_identifier.decode().strip() or None,
            "creation_date": date_convert(pvd.volume_creation_date),
            "expiration_date": date_convert(pvd.volume_expiration_date),
            "effective_date": date_convert(pvd.volume_effective_date),
            "escape_seq": f"00 * {len(pvd.escape_sequences)}" if pvd.escape_sequences == bytearray(
                len(pvd.escape_sequences)) else pvd.escape_sequences,
            "set_size": pvd.set_size,
            "seq_num": pvd.seqnum,
            "path_tbl_size": pvd.path_tbl_size,
            "path_table_location_le": pvd.path_table_location_le,
            "path_table_location_be": pvd.path_table_location_be,
            "optional_path_table_location_le": pvd.optional_path_table_location_le,
            "optional_path_table_location_be": pvd.optional_path_table_location_be,
            "application_reserve": None if pvd.application_use == bytearray(512) else pvd.application_use
        }
        g.LOG.write(f"Got Primary Volume Descriptor: {pvd}\n")
        if js:
            # cefpython complaints it's too big for an `int`
            pvd["size"] = str(pvd["size"])
            js.Call(pvd)
        return pvd

    def get_files(self, path="/", no_versions=True):
        """
        Read and list file paths directly from the disc device file system
        which doesn't require the device to be mounted

        Returns a tuple generator of the file path which will be
        absolute-paths relative to the root of the device, the Logical
        Block Address (LBA), and the Size (in sectors).
        """
        for child in self.cdlib.list_children(iso_path=path):
            file_path = child.file_identifier().decode()
            # skip the `.` and `..` paths
            if file_path in [".", ".."]:
                continue
            # remove the semicolon and version number
            if no_versions and ";" in file_path:
                file_path = file_path.split(";")[0]
            # join it to root to be absolute
            file_path = os.path.join("/", path, file_path)
            # get lba
            lba = child.extent_location()
            # get size in sectors
            size = int(child.get_data_length() / self.dvdcss.SECTOR_SIZE)
            g.LOG.write(f"Found title file: {file_path}, lba: {lba}, size: {size}")
            yield file_path, lba, size

    def get_vob_lbas(self, crack_keys=False):
        """
        Get the LBA data for all VOB files in disc.
        Optionally seek with SEEK_KEY flag to obtain keys.

        Raises SlipstreamSeekError on seek failures.
        """
        # Create an array for holding the title data
        lba_data = []
        # Loop all files in disc:/VIDEO_TS
        for vob, lba, size in self.get_files("/VIDEO_TS"):
            # we only want vob files
            if os.path.splitext(vob)[-1] != ".VOB":
                continue
            # get title key
            if crack_keys:
                if lba == self.dvdcss.seek(lba, self.dvdcss.SEEK_KEY):
                    g.LOG.write(f"Got title key for {vob}")
                else:
                    raise SlipstreamSeekError(
                        f"Failed to seek the disc to {lba} while attempting to "
                        f"crack the title key for {os.path.basename(vob)}"
                    )
            # add data to title offsets
            lba_data.append((lba, size))
        # Return lba data
        return lba_data

    @asynchronous_auto
    def create_backup(self, js=None):
        """
        Create a full untouched (but decrypted) ISO backup of a DVD with all
        metadata intact.

        Raises SlipstreamNoKeysObtained if no CSS keys were obtained when needed.
        Raises SlipstreamReadError on unexpected read errors.
        """
        try:
            # Notify JS-land we're starting
            if js:
                js.Call(True)
            # Print primary volume descriptor information
            g.LOG.write(f"Starting DVD backup for {self.dev}")
            pvd = self.cdlib.pvds[0]
            pvd.volume_identifier = pvd.volume_identifier.decode().strip()
            fn = f"{pvd.volume_identifier}.ISO"
            fn_tmp = f"{pvd.volume_identifier}.ISO.tmp"
            first_lba = 0
            last_lba = pvd.space_size - 1
            disc_size = pvd.space_size * self.dvdcss.SECTOR_SIZE
            g.LOG.write(
                f"Reading sectors {first_lba:,} to {last_lba:,} with sector size {self.dvdcss.SECTOR_SIZE:,} B.\n"
                f"Length: {last_lba + 1:,} sectors, {disc_size:,} bytes.\n"
                f'Saving to "{fn}"...'
            )
            # Retrieve CSS keys if disc is scrambled
            if self.dvdcss.is_scrambled():
                g.LOG.write("DVD is scrambled. Checking if all CSS keys can be cracked. This might take a while.")
                self.vob_lba_offsets = self.get_vob_lbas(crack_keys=True)
                if not self.vob_lba_offsets:
                    raise SlipstreamNoKeysObtained("No CSS title keys were returned, unable to decrypt.")
            else:
                g.LOG.write("DVD isn't scrambled. CSS title key cracking skipped.")
            # Create a file write handle to temp file
            f = open(fn_tmp, "wb")
            # Create a TQDM progress bar
            t = tqdm(total=last_lba + 1, unit="sectors", file=TqdmHook())
            # Read through all the sectors in a memory efficient manner
            current_lba = first_lba
            g.LOG.write(f"Reading sectors {current_lba}->{last_lba}...")
            while current_lba <= last_lba:
                # get the maximum sectors to read at once
                sectors = min(self.dvdcss.BLOCK_BUFFER, last_lba - current_lba + 1)
                # read sectors
                read_sectors = self.read(current_lba, sectors)
                if read_sectors < 0:
                    raise SlipstreamReadError(f"An unexpected read error occurred reading {current_lba}->{sectors}")
                # write the buffer to output file
                f.write(self.dvdcss.buffer)
                # increment the current sector and update the tqdm progress bar
                current_lba += read_sectors
                # write progress to GUI log
                g.PROGRESS.c.Call((current_lba / last_lba) * 100)
                # write progress to CLI log
                t.update(read_sectors)
            # Close file and tqdm progress bar
            f.close()
            t.close()
            # Rename temp file to final filename
            os.rename(fn_tmp, fn)
            # Tell the user some output information
            g.LOG.write(
                "Finished DVD Backup!\n"
                f"Read a total of {current_lba:,} sectors ({os.path.getsize(fn):,}) bytes.\n"
            )
        finally:
            # Notify js-land were done
            if js:
                js.Call(False)

    def read(self, first_lba, sectors):
        """
        Efficiently read an amount of sectors from the disc while supporting decryption
        with libdvdcss (pydvdcss).

        Returns the amount of sectors read.
        Raises a SlipstreamSeekError on Seek Failures and SlipstreamReadError on Read Failures.
        """

        # we need to seek to the first sector. Otherwise we get faulty data.
        needToSeek = first_lba != self.reader_position or first_lba == 0
        inTitle = False
        enteredTitle = False

        # Make sure we never read encrypted and unencrypted data at once since libdvdcss
        # only decrypts the whole area of read sectors or nothing at all.
        for vob_lba_offset in self.vob_lba_offsets:
            titleStart = vob_lba_offset[0]
            titleEnd = titleStart + vob_lba_offset[1] - 1

            # update key when entering a new title
            # FIXME: we also need this if we seek into a new title (not only the start of the title)
            if titleStart == first_lba:
                enteredTitle = needToSeek = inTitle = True

            # if first_lba < titleStart and first_lba + sectors > titleStart:
            if first_lba < titleStart < first_lba + sectors:
                # read range will read beyond or on a title,
                # let's read up to right before the next title start
                sectors = titleStart - first_lba

            # if first_lba < titleEnd and first_lba + sectors > titleEnd:
            if first_lba < titleEnd < first_lba + sectors:
                # read range will read beyond or on a title,
                # let's read up to right before the next title start
                sectors = titleEnd - first_lba + 1

            # is our read range part of one title
            if first_lba >= titleStart and first_lba + (sectors - 1) <= titleEnd:
                inTitle = True

        if needToSeek:
            flags = self.dvdcss.NOFLAGS
            if enteredTitle:
                flags = self.dvdcss.SEEK_KEY
            elif inTitle:
                flags = self.dvdcss.SEEK_MPEG

            # refresh the key status for this sector's data
            self.reader_position = self.dvdcss.seek(first_lba, flags)
            if self.reader_position != first_lba:
                raise SlipstreamSeekError(f"Failed to seek the disc to {first_lba} while doing a device read.")

        flags = self.dvdcss.NOFLAGS
        if inTitle:
            flags = self.dvdcss.READ_DECRYPT

        ret = self.dvdcss.read(sectors, flags)
        if ret != sectors:
            raise SlipstreamReadError(f"An unexpected read error occurred reading {first_lba}->{first_lba + sectors}")
        self.reader_position += ret

        return ret


class TqdmHook:
    """hook to simply intercept tqdm's progress messages and log them."""

    @staticmethod
    def write(text):
        # log the tqdm progress message
        g.LOG.write(text, echo=False)
        # return it to stdout
        return sys.stdout.write(text)

    @staticmethod
    def flush():
        # return it to stdout, we don't need to flush the log
        sys.stdout.flush()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Slipstream - The most informative Home-media backup solution.
Copyright (C) 2020 PHOENiX

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

~~~

Configuration object used to manage the loading and saving of
configuration entries. It deals with everything including the reading and
writing of yaml files, ensuring paths exist, ensuring it has read and write
permissions and such.
"""

import os.path

import yaml


class Config(object):
    def __init__(self, config_path):
        self.settings = {
            "-": "-"  # need at least one config entry, otherwise it infinite loops
            # todo ; implement stuff that needs configuration
        }
        self.config_path = config_path

    def get_handle(self, mode="r"):
        """Open a file handle with the specified mode"""
        os.makedirs(os.path.dirname(self.config_path), exist_ok=True)
        return open(self.config_path, mode)

    def load(self):
        """Load yaml config file as a dictionary"""
        if (
            not os.path.exists(self.config_path) or
            not os.path.isfile(self.config_path) or
            not os.path.exists(self.config_path)
        ):
            # no config file exists yet, let's create base one
            self.save()
        with self.get_handle() as f:
            stored_settings = yaml.safe_load(f)
            if not stored_settings:
                # invalid yaml or empty file, reset the config file and reload
                self.save()
                self.load()
                return
            self.settings = stored_settings

    def save(self, settings=None):
        """Save yaml config to file from dictionary"""
        if settings:
            self.settings = settings
        if self.settings:
            with self.get_handle("w") as f:
                yaml.dump(self.settings, f)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Slipstream - The most informative Home-media backup solution.
Copyright (C) 2020 PHOENiX

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""


class KeyboardHandler:

    # Important: Functions cannot be static!

    def OnKeyEvent(self, browser, event, **_):
        """Called after the renderer and javascript in the page has had a chance to handle the event."""
        if event["type"] == 3:
            # CTRL+SHIFT+I - Open Dev Tools
            if event["modifiers"] == 6 and event["native_key_code"] == 31:
                browser.ShowDevTools()



import os
import sys
from shutil import rmtree

import pslipstream.cfg as cfg


def print_bold(s):
    """Prints things in bold."""
    print("\033[1m{0}\033[0m".format(s))


def build_clean():
    print_bold("Removing previous builds…")
    try:
        rmtree(os.path.join(cfg.root_dir, "dist"))
    except OSError:
        pass
    try:
        rmtree(os.path.join(cfg.root_dir, "build"))
    except OSError:
        pass


def build():
    build_clean()
    print_bold("Ensuring an up-to-date environment…")
    os.system("{0} -m pip install --user --upgrade setuptools wheel".format(sys.executable))
    print_bold("Building Source and Wheel (universal) distribution…")
    os.system("{0} setup.py sdist bdist_wheel --universal".format(sys.executable))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Slipstream - The most informative Home-media backup solution.
Copyright (C) 2020 PHOENiX

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import argparse
import builtins as g
import hashlib
import os
import webbrowser

import requests
from appdirs import user_data_dir

from cefpython3 import cefpython as cef

import pslipstream.cfg as cfg
from pslipstream.config import Config
from pslipstream.dvd import Dvd
from pslipstream.gui import Gui
from pslipstream.helpers import get_device_list
from pslipstream.log import Log
from pslipstream.progress import Progress


def main():
    # Prepare Metadata
    cfg.cef_version = cef.GetVersion()
    cfg.user_dir = user_data_dir(cfg.title_pkg, cfg.author)
    cfg.config_file = os.path.join(cfg.user_dir, "config.yml")

    # Initialize custom global variables
    g.ARGS = get_arguments()
    g.LOG = Log()  # Logger, everything written here gets print()'d and sent to GUI
    g.PROGRESS = Progress()  # Progress Bar, controls only the GUI's progress bar.
    g.DBG = g.ARGS.dbg  # Debug switch, enables debugging specific code and logging
    g.CFG = Config(cfg.config_file)
    g.CFG.load()

    # Print License if asked
    if g.ARGS.license:
        if not os.path.exists("LICENSE"):
            # download the license from the official source if not found locally
            lic_url = "https://www.gnu.org/licenses/gpl-3.0.txt"
            lic_text = requests.get(lic_url).text
            # noinspection SpellCheckingInspection
            expected_md5 = "1ebbd3e34237af26da5dc08a4e440465"
            computed_md5 = hashlib.md5(lic_text.encode("utf-8")).hexdigest()
            if computed_md5 != expected_md5:
                print(
                    "License file was not found locally and it also failed to download :(\n"
                    f"If you would like to read the LICENSE file, head over to {lic_url}"
                )
            else:
                print(lic_text)
        else:
            with open("LICENSE", mode="rt", encoding="utf-8") as f:
                print(f.read())
        exit(0)

    # Get and Print Runtime Details
    g.LOG.write(get_runtime_details() + "\n")

    # Let's get to it
    if g.ARGS.cli:
        cli()
    else:
        gui()


def get_arguments():
    ap = argparse.ArgumentParser()
    ap.add_argument(
        "--dev",
        nargs="*",
        required=False,
        help="This is only to be used by dev's that know what it does!!",
    )
    ap.add_argument(
        "--dbg",
        action="store_true",
        default=False,
        required=False,
        help="Debug mode, may have increased logging that could hurt performance.",
    )
    ap.add_argument(
        "--license",
        action="store_true",
        default=False,
        required=False,
        help="View license details",
    )
    ap.add_argument(
        "--cli",
        action="store_true",
        default=False,
        required=False,
        help="Setting this stops the GUI from running",
    )
    ap.add_argument(
        "-d",
        "--device",
        type=str,
        default="",
        required=False,
        help="Choose device for backup (e.g. '/dev/sr0', '/mnt/dvd-rw', 'E:')",
    )
    return ap.parse_args()


def get_runtime_details():
    return "\n".join(
        [
            cfg.copyright_paragraph,
            "",
            f"{cfg.title} v{cfg.version}",
            cfg.description,
            cfg.url,
            "",
            f":: {'DEBUG' if g.DBG else 'Standard'} MODE",
            f":: {cfg.platform} {cfg.architecture} (Python v{cfg.py_version})",
            f":: CEF Runtime: {cfg.cef_version}",
            f":: User Directory: {cfg.user_dir}",
            f":: Static Directory: {cfg.static_dir}",
        ]
    )


def gui():
    # set ui index location based on environment
    if g.ARGS.dev:
        port = None
        if "ui-build" in g.ARGS.dev:
            # gatsby build && gatsby serve
            port = 9000
        elif "ui-develop" in g.ARGS.dev:
            # gatsby develop
            port = 8000
        cfg.ui_index = "http://localhost:" + str(port)
    else:
        cfg.ui_index = "https://slipstream-ui.vercel.app"
    # create gui, and fire it up
    g.GUI = Gui(
        url=cfg.ui_index,
        icon=cfg.icon_file,
        js_bindings={
            "properties": [
                {"name": "config", "item": g.CFG.settings},
            ],
            "objects": [
                {"name": "dvd", "item": Dvd()},
                {"name": "log", "item": g.LOG},
                {"name": "progress", "item": g.PROGRESS},
            ],
            "functions": [
                {"name": "pyDelete", "item": os.remove},
                {"name": "pyHref", "item": lambda url: webbrowser.open(url)},
                {"name": "configSave", "item": g.CFG.save},
                {"name": "getDeviceList", "item": get_device_list},
            ],
        },
    )
    g.GUI.mainloop()


def cli():
    d = Dvd()
    d.open(g.ARGS.device)
    d.create_backup()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Slipstream - The most informative Home-media backup solution.
Copyright (C) 2020 PHOENiX

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

~~~

Definitions of Slipstream exceptions.
"""


class TkinterVersionError(Exception):
    """Tkinter version is too outdated, update tkinter to continue."""


class CefInitializationError(Exception):
    """CEF Failed to initialize."""


class WindowHandleError(Exception):
    """Couldn't obtain the GUI's window handle."""


class SlipstreamUiError(Exception):
    """Failed to load the UI."""


class SlipstreamDiscInUse(Exception):
    """A disc is already initialised in this instance."""


class SlipstreamNoKeysObtained(Exception):
    """No keys were returned, unable to decrypt."""


class SlipstreamReadError(Exception):
    """An unexpected read error occurred."""


class SlipstreamSeekError(Exception):
    """An unexpected seek error occurred."""



class Progress:
    def __init__(self):
        self.progress = 0
        self.c = None

    def set_c(self, js):
        # todo ; rename function to set_js_callback to be more descriptive
        self.c = js



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Slipstream - The most informative Home-media backup solution.
Copyright (C) 2020 PHOENiX

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import builtins as g
import webbrowser

from pslipstream.exceptions import SlipstreamUiError


class LoadHandler:

    # Important: Functions cannot be static!

    def OnLoadingStateChange(self, browser, is_loading, **_):
        """Called when the loading state has changed."""
        if browser.GetIdentifier() != 1:
            return  # this is a href hook, lets just pass
        if not is_loading:
            # browser dom ready
            # ...
            pass

    def OnLoadStart(self, browser, frame):
        """Called when loading starts."""
        # hook and open in users browser instead of inside App
        if browser.GetIdentifier() != 1:
            browser.StopLoad()
            browser.CloseBrowser()
            webbrowser.open(frame.GetUrl())

    def OnLoadError(self, browser, **_):
        """Called when loading errors."""
        if browser.GetIdentifier() != 1:
            return  # this is a href, lets just pass
        if "gatsby-develop" in g.ARGS.dev:
            # gatsby-develop switch needs this to pass, I believe it's caused by the 404 page check
            # failing, and because it failed, OnLoadError gets called.
            return
        raise SlipstreamUiError()



import os
import subprocess
import sys

from setuptools import Command

import pslipstream.cfg as cfg
from setup_commands import build_clean, print_bold


class PackCommand(Command):
    """Support setup.py pack."""

    description = "Pack the repo with PyInstaller."
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    @staticmethod
    def run():
        from PyInstaller.utils.hooks import get_package_paths
        build_clean()
        print_bold("Ensuring supported environment…")
        if not cfg.windows and not cfg.linux and not cfg.darwin:
            print("Sorry! Only Windows, Linux and Darwin platforms are supported.")
            sys.exit(1)
        print_bold("Ensuring PyInstaller is available…")
        os.system(
            "{0} -m pip install --user --upgrade pyinstaller".format(sys.executable)
        )
        print_bold("Packing with PyInstaller…")
        sep = ";" if cfg.windows else ":"
        sub = subprocess.Popen(
            [
                "pyinstaller",
                "--clean",
                "-F",
                f"{cfg.title_pkg}/__init__.py",
                "--add-data",
                f"{cfg.title_pkg}/static{sep}static",
                "--add-data",
                get_package_paths("cefpython3")[1].replace("\\", "/")
                + f"{sep}.",
                "--hidden-import",
                "pkg_resources.py2_warn",
                "-n",
                "Slipstream",
            ]
        )
        sub.communicate()
        if sub.returncode != 0:
            print("Oh no! PyInstaller failed, code=%s" % sub.returncode)
            build_clean()
            sys.exit(1)
        sys.exit()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Slipstream - The most informative Home-media backup solution.
Copyright (C) 2020 PHOENiX

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

~~~

Metadata used throughout the application.
By calculating all the metadata in meta.py, it allows us
to easily expose the information to the user via import as well as
throughout the code by importing only what we want.
"""

import datetime
import os
import platform

import pkg_resources

# general
title = "Slipstream"
title_pkg = "pslipstream"
description = "The most informative Home-media backup solution."
url = "https://github.com/rlaPHOENiX/Slipstream"
version = "0.4.0"
author = "PHOENiX"
author_email = "rlaphoenix@pm.me"
min_size = "1200x440"
package_obj = None
try:
    package_obj = pkg_resources.Requirement.parse(f"{title_pkg}=={version}")
except pkg_resources.DistributionNotFound:
    pass

# build configuration
py_ver_support = ">=3.6, <3.8"
# noinspection SpellCheckingInspection
req_packages = [
    # general
    "appdirs>=1.4.4",
    "tqdm>=4.46.1",
    # cef
    "cefpython3==66.0",
    "pyobjc; sys_platform == 'darwin'",
    "AppKit; sys_platform == 'darwin'",
    # http and such
    "requests>=2.24.0",
    # parsing, syntax, and validators
    "PyYAML>=5.3.1",
    "python-dateutil>=2.8.1",
    # general disc operations
    "pycdlib>=1.10.0",
    # dvd disc operations
    "pydvdcss>=1.0.7.post0",
    "rlapydvdid>=1.2"  # fork of pydvdid, faster and doesn't need mount permission
]
# noinspection SpellCheckingInspection
opt_packages = {
    # build related
    "packing support": ["PyInstaller"]  # provide no version to let the user decide
}

# environment
cef_version = None  # gotten on main()
py_version = platform.python_version()
architecture = platform.architecture()[0]
platform = platform.system()
windows = platform == "Windows"
linux = platform == "Linux"
darwin = platform == "Darwin"

# licensing and copyright
licence = "GPLv3"
copyright_line = f"Copyright (C) {datetime.datetime.now().year} {author}"
copyright_paragraph = "\n".join([
    f"{title}  {copyright_line}",
    "This program comes with ABSOLUTELY NO WARRANTY.",
    "This is free software, and you are welcome to redistribute it",
    f"under certain conditions; type '{title_pkg} --license' for details."
])

# directories
root_dir = os.path.dirname(os.path.abspath(os.path.dirname(__file__)))
user_dir = None  # gotten on main()
static_dir = os.path.join(root_dir, "static")
if package_obj:
    try:
        static_dir = pkg_resources.resource_filename(package_obj, f"{title_pkg}/static")
    except pkg_resources.DistributionNotFound:
        pass
    except pkg_resources.VersionConflict:
        pass

# file paths
config_file = None  # gotten on main()
icon_file = os.path.join(static_dir, "icon.png")
ui_index = None  # prefix with `file://` for local file



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Slipstream - The most informative Home-media backup solution.
Copyright (C) 2020 PHOENiX

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import ctypes
import platform
import tkinter as tk

from cefpython3 import cefpython as cef

import pslipstream.exceptions as exceptions
from pslipstream.gui.keyboard_handler import KeyboardHandler
from pslipstream.gui.load_handler import LoadHandler


class BrowserFrame(tk.Frame):

    def __init__(self, master, url, js_bindings, on_load=None, on_hot_key=None):
        self.js_bindings = js_bindings
        self.platform = platform.system()
        self.url = url
        self.on_load = on_load
        self.on_hot_key = on_hot_key
        self.closing = False
        self.browser = None
        tk.Frame.__init__(self, master)
        self.bind("<FocusIn>", self.on_focus_in)
        self.bind("<FocusOut>", self.on_focus_out)
        self.bind("<Configure>", self.on_configure)
        self.focus_set()

    def embed_browser(self):
        window_info = cef.WindowInfo()
        window_info.SetAsChild(self.get_window_handle(), [0, 0, self.winfo_width(), self.winfo_height()])
        self.browser = cef.CreateBrowserSync(url=self.url, window_info=window_info)
        if self.platform == "Windows":
            window_handle = self.browser.GetOuterWindowHandle()
            insert_after_handle = 0
            SWP_NO_MOVE = 0x0002  # ignore x, y params (don't move to 0, 0)
            ctypes.windll.user32.SetWindowPos(window_handle, insert_after_handle, 0, 0, 1300, 440, SWP_NO_MOVE)
        # Set Handlers
        if self.js_bindings:
            js_bindings_ = cef.JavascriptBindings(bindToFrames=False, bindToPopups=False)
            if "properties" in self.js_bindings:
                for property_ in self.js_bindings["properties"]:
                    js_bindings_.SetProperty(property_["name"], property_["item"])
            if "objects" in self.js_bindings:
                for object_ in self.js_bindings["objects"]:
                    js_bindings_.SetObject(object_["name"], object_["item"])
            if "functions" in self.js_bindings:
                for function_ in self.js_bindings["functions"]:
                    js_bindings_.SetFunction(function_["name"], function_["item"])
            self.browser.SetJavascriptBindings(js_bindings_)
        self.browser.SetClientHandler(KeyboardHandler())
        self.browser.SetClientHandler(LoadHandler())
        self.message_loop_work()

    def get_window_handle(self):
        if self.winfo_id() > 0:
            return self.winfo_id()
        elif self.platform == "Darwin":
            # On Mac window id is an invalid negative value (Issue #308).
            # This is kind of a dirty hack to get window handle using
            # PyObjC package. If you change structure of windows then you
            # need to do modifications here as well.
            # noinspection PyUnresolvedReferences
            from AppKit import NSApp
            # noinspection PyUnresolvedReferences
            import objc
            # Sometimes there is more than one window, when application
            # didn't close cleanly last time Python displays an NSAlert
            # window asking whether to Reopen that window.
            # noinspection PyUnresolvedReferences
            return objc.pyobjc_id(NSApp.windows()[-1].contentView())
        else:
            raise exceptions.WindowHandleError()

    def message_loop_work(self):
        cef.MessageLoopWork()
        self.after(10, self.message_loop_work)

    def on_configure(self, _):
        if not self.browser:
            self.embed_browser()

    def on_root_configure(self):
        # Root <Configure> event will be called when top window is moved
        if self.browser:
            self.browser.NotifyMoveOrResizeStarted()

    def on_mainframe_configure(self, width, height):
        if self.browser:
            if self.platform == "Windows":
                ctypes.windll.user32.SetWindowPos(self.browser.GetWindowHandle(), 0, 0, 0, width, height, 0x0002)
            elif self.platform == "Linux":
                self.browser.SetBounds(0, 0, width, height)
            self.browser.NotifyMoveOrResizeStarted()

    def on_focus_in(self, _):
        if self.browser:
            self.browser.SetFocus(True)

    def on_focus_out(self, _):
        if self.browser:
            self.browser.SetFocus(False)

    def on_root_close(self):
        if self.browser:
            self.browser.CloseBrowser(True)
            self.clear_browser_references()
        self.destroy()

    def clear_browser_references(self):
        # Clear browser references that you keep anywhere in your
        # code. All references must be cleared for CEF to shutdown cleanly.
        self.browser = None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Greets to https://github.com/navdeep-G/setup.py

# Note: To use the 'upload' functionality of this file, you must:
#   $ pipenv install twine --dev

# std
import io
import os

# pip packages
from setuptools import find_packages, setup

# slipstream
import pslipstream.cfg as cfg
from setup_commands.dist import DistCommand
from setup_commands.pack import PackCommand

# Import the README and use it as the long-description.
try:
    with io.open(os.path.join(cfg.root_dir, "README.md"), encoding="utf-8") as f:
        long_description = ("\n" + f.read(), "text/markdown")
except FileNotFoundError:
    long_description = (cfg.description, "text/plain")

# Where the magic happens:
setup(
    name=cfg.title_pkg,
    version=cfg.version,
    description=cfg.description,
    long_description=long_description[0],
    long_description_content_type=long_description[1],
    author=cfg.author,
    author_email=cfg.author_email,
    python_requires=cfg.py_ver_support,
    url=cfg.url,
    project_urls={
        # todo ; "Documentation": '...',
        "Source": "https://github.com/rlaPHOENiX/Slipstream",
    },
    packages=find_packages(),
    entry_points={
        "console_scripts": [f"{cfg.title_pkg}={cfg.title_pkg}.slipstream:main"]
    },
    install_requires=cfg.req_packages,
    extras_require=cfg.opt_packages,
    include_package_data=True,
    package_data={
        "": ["LICENSE", "README.md", "HISTORY.md", "static/*"]
    },
    license=cfg.licence,
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Development Status :: 3 - Alpha",
        "Natural Language :: English",
        "Intended Audience :: End Users/Desktop",
        "Operating System :: OS Independent",
        "Topic :: Multimedia",
        "Topic :: Multimedia :: Video",
        "Topic :: Multimedia :: Video :: Conversion",
    ],
    # $ setup.py publish support.
    cmdclass={"dist": DistCommand, "pack": PackCommand},
)
